/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.util.LinkedHashMap;
import java.util.Map;

import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;

import org.skife.jdbi.v2.Handle;
import org.skife.jdbi.v2.IDBI;
import org.skife.jdbi.v2.tweak.HandleCallback;

import com.akiban.rest.dbo.JdbcDataSource;
import com.akiban.rest.util.EncodingUtil;
import com.google.inject.Inject;

/**
 * Implementation of REST-oriented Get, Multi-Get, Create, Update, Delete and
 * Multi-Delete.
 * 
 * FIXME : refactor mutation methods to use Transaction and handle rollbacks
 * properly on intermediate errors
 */
@Path("/{schema}.{table}")
public class DataAccessOperationsResource {
	@Inject
	private IDBI database;

	@Inject
	private JdbcDataSource store;

	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public Response retrieveEntity(@QueryParam("format") String format,
			@QueryParam("jsoncallback") String jsonp,
			@PathParam("schema") String schema,
			@PathParam("table") String table,
			@QueryParam("depth") Integer depth,
			@QueryParam("offset") Integer offset,
			@QueryParam("limit") Integer limit) throws Exception {
		return store.retrieve(format, jsonp, schema, table, "", depth, offset,
				limit);
	}

	@GET
	@Path("{identifiers:.*}")
	@Produces(MediaType.APPLICATION_JSON)
	public Response retrieveEntity(@QueryParam("format") String format,
			@QueryParam("jsoncallback") String jsonp,
			@PathParam("schema") String schema,
			@PathParam("table") String table,
			@QueryParam("depth") Integer depth, @Context UriInfo uri)
			throws Exception {
		return store.retrieve(format, jsonp, schema, table,
				getPrimaryKeys(uri), depth, null, null);
	}

	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	@Produces(MediaType.APPLICATION_JSON)
	public Response createEntity(@PathParam("schema") final String schema,
			@PathParam("table") final String table, final byte[] entityBytes)
			throws Exception {
		final Map<String, Object> entity = EncodingUtil.fromJson(
				LinkedHashMap.class, entityBytes);

		return database.withHandle(new HandleCallback<Response>() {
			@Override
			public Response withHandle(Handle handle) throws Exception {
				return store.insert(handle, schema, table, entity);
			}
		});
	}

	@PUT
	@Path("{identifiers:.*}")
	@Produces(MediaType.APPLICATION_JSON)
	public Response updateEntity(@PathParam("schema") final String schema,
			@PathParam("table") final String table, final byte[] entityBytes,
			@Context final UriInfo uri) throws Exception {
		final Map<String, Object> entity = EncodingUtil.fromJson(
				LinkedHashMap.class, entityBytes);

		return database.withHandle(new HandleCallback<Response>() {
			@Override
			public Response withHandle(Handle handle) throws Exception {
				return store.update(handle, schema, table, getPrimaryKeys(uri),
						entity);
			}
		});
	}

	@DELETE
	@Path("{identifiers:.*}")
	@Produces(MediaType.APPLICATION_JSON)
	public Response deleteEntity(@PathParam("schema") String schema,
			@PathParam("table") String table, @Context UriInfo uri)
			throws Exception {
		return store.delete(schema, table, getPrimaryKeys(uri));
	}

	private static String getPrimaryKeys(UriInfo uri) {
		//
		// FIXME: bulletproof the split() below
		//
		return uri.getPath().split("/")[1];
	}
}
