/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.ContextHandler;
import org.eclipse.jetty.server.handler.HandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.util.thread.QueuedThreadPool;

import com.akiban.rest.util.BasicConfigurationService;
import com.akiban.rest.util.Props;
import com.akiban.server.plugins.httpconductor.HttpConductor;
import com.akiban.server.service.Service;
import com.akiban.server.service.config.ConfigurationService;
import com.google.inject.AbstractModule;
import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.name.Names;

/**
 * Starts the embedded Jetty server.
 */
public class Main implements HttpConductor, Service {
	private static volatile Injector injector;
	private final Props props;
	private final Server server;

	public static void main(String[] args) {
		if (injector == null) {
			injector = Guice.createInjector(new EmbeddedRestModule(),
					new RestServiceModule());
		}

		injector.getInstance(RestServiceImpl.class).start();
		injector.getInstance(Main.class).start();
	}

	protected static void setInjector(Injector inject) {
		injector = inject;
	}

	protected static Injector getInjector() {
		return injector;
	}

	@Inject
	public Main(Props config, Server jetty) {
		this.props = config;
		this.server = jetty;

		if (System.getProperty("java.net.preferIPv4Stack") == null) {
			System.setProperty("java.net.preferIPv4Stack", "true");
		}

		SelectChannelConnector connector = new SelectChannelConnector();

		connector.setHost(props.getString("http.host"));
		connector.setPort(props.getInt("http.port"));
		connector.setThreadPool(new QueuedThreadPool(props
				.getInt("http.threads")));
		connector.setAcceptors(props.getInt("http.acceptors"));
		connector.setMaxIdleTime(props.getInt("http.max_idle_time"));
		connector.setAcceptQueueSize(props.getInt("http.accept_queue_size"));
		connector.setLowResourcesConnections(props
				.getInt("http.low_resources_connections"));

		server.setConnectors(new Connector[] { connector });
	}

	@Override
	public void registerHandler(ContextHandler handler) {
		HandlerCollection handlers = new HandlerCollection();
		handlers.setHandlers(new Handler[] { handler });

		server.setHandler(handlers);
	}

	@Override
	public void unregisterHandler(ContextHandler handler) {
		HandlerCollection handlers = new HandlerCollection();
		handlers.setHandlers(new Handler[] {});

		server.setHandler(handlers);
	}

	@Override
	public void start() {
		try {
			server.start();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void stop() {
		try {
			server.stop();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void crash() {
		throw new UnsupportedOperationException();
	}

	public static class EmbeddedRestModule extends AbstractModule {
		@Override
		protected void configure() {
			bind(String.class).annotatedWith(
					Names.named("com.akiban.rest.config.resource")).toInstance(
					"com/akiban/server/plugin-configuration.properties");
			bind(ConfigurationService.class)
					.to(BasicConfigurationService.class).asEagerSingleton();
			bind(Main.class).asEagerSingleton();
			bind(Server.class).asEagerSingleton();
			bind(HttpConductor.class).to(Main.class);
			bind(RestService.class).to(RestServiceImpl.class)
					.asEagerSingleton();
		}
	}
}
