/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.util.EnumSet;

import javax.servlet.DispatcherType;

import org.eclipse.jetty.servlet.ServletContextHandler;
import org.postgresql.Driver;

import com.akiban.rest.util.EmptyServlet;
import com.akiban.rest.util.Props;
import com.akiban.server.plugins.httpconductor.HttpConductor;
import com.akiban.server.service.Service;
import com.google.inject.Inject;
import com.google.inject.servlet.GuiceFilter;
import com.jolbox.bonecp.BoneCPDataSource;

/**
 * RestService plugin implementation, compliant with
 * com.akiban.server.service.Service.
 */
public class RestServiceImpl implements RestService, Service {
	private final Props props;
	private final HttpConductor http;
	private final BoneCPDataSource dataSource;
	private volatile ServletContextHandler handler;

	@Inject
	public RestServiceImpl(Props props, HttpConductor http,
			BoneCPDataSource dataSource) {
		this.props = props;
		this.http = http;
		this.dataSource = dataSource;
	}

	@Override
	public void start() {
		if (this.handler == null) {
			configureDataSource(dataSource, props);
		}
		registerConnector(http);
	}

	@Override
	public void stop() {
	}

	@Override
	public void crash() {
		throw new UnsupportedOperationException();
	}

	private void registerConnector(HttpConductor http) {
		if (this.handler == null) {
			ServletContextHandler context = new ServletContextHandler();

			context.setContextPath(props.getString("http.context_path"));
			context.addFilter(GuiceFilter.class, "/*",
					EnumSet.<DispatcherType> of(DispatcherType.REQUEST));
			context.addServlet(EmptyServlet.class, "/*");

			this.handler = context;
		}

		http.registerHandler(this.handler);
	}

	public static void configureDataSource(BoneCPDataSource dataSource,
			Props props) {
		dataSource.setDriverClass(Driver.class.getName());
		dataSource.setJdbcUrl(props.getString("jdbc.url"));
		dataSource.setUsername(props.getString("jdbc.user"));
		dataSource.setPassword(props.getString("jdbc.password"));
		dataSource
				.setMinConnectionsPerPartition(props.getInt("jdbc.min_conns"));
		dataSource
				.setMaxConnectionsPerPartition(props.getInt("jdbc.max_conns"));
		dataSource.setAcquireRetryDelayInMs(props
				.getInt("jdbc.conn_retry_delay_ms"));
		dataSource.setConnectionTestStatement("select 1");
		dataSource.setInitSQL("set OutputFormat = 'json'");
	}
}
