/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.dbo;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;

import org.skife.jdbi.v2.Handle;

import com.akiban.rest.model.Entity;
import com.akiban.rest.model.GroupInfo;
import com.akiban.rest.model.PrimaryKeyInfo;
import com.akiban.rest.model.PseudoPk;
import com.google.inject.Inject;

/**
 * Helper methods for representing JSON objects which have been provided by the
 * user for entry into the database.
 */
public class EntityHelper {
	@Inject
	private SqlGroupHelper groupHelper;

	@Inject
	private SqlPrimaryKeyHelper pkHelper;

	public List<Entity> parseEntities(Handle handle, String schema,
			String table, Map<String, Object> attributes,
			List<GroupInfo> groupInfos) throws Exception {
		List<Entity> results = new ArrayList<Entity>();

		collectEntities(handle, schema, table, attributes, results,
				Collections.<String, Entity> emptyMap(), new AtomicLong());

		return results;
	}

	private void collectEntities(Handle handle, String schema, String table,
			Map<String, Object> attributes, List<Entity> results,
			Map<String, Entity> enclosingEntities, AtomicLong seq)
			throws Exception {
		PrimaryKeyInfo pkInfo = pkHelper.getPrimaryKeyInfo(handle, schema,
				table);
		final List<GroupInfo> parentGroupInfos = groupHelper
				.getParentGroupInfo(handle, schema, table);

		if (!pkInfo.hasPrimaryKey(attributes) && pkInfo.isIdentitySequence()) {
			attributes.put(pkInfo.getColumns().get(0), new PseudoPk(seq));
		}

		final List<GroupInfo> childGroupInfos = groupHelper.getChildGroupInfo(
				handle, schema, table);

		Map<String, Object> newAttributes = new LinkedHashMap<String, Object>();
		newAttributes.putAll(attributes);

		for (GroupInfo group : childGroupInfos) {
			if (!newAttributes.containsKey(group.getRightKey())
					|| newAttributes.get(group.getRightKey()) == null) {
				String parentKey = SqlHelper.getTableName(
						group.getParentSchema(), group.getParentTable());
				Entity parentEntity = enclosingEntities.get(parentKey);

				if (parentEntity != null) {
					newAttributes.put(group.getRightKey(), parentEntity
							.getAttributes().get(group.getLeftKey()));
				}
			}
		}

		Entity thisEntity = new Entity(schema, table, pkInfo, newAttributes,
				enclosingEntities);
		results.add(thisEntity);

		for (GroupInfo group : parentGroupInfos) {
			String nestedKey = group.getChildSchema() + "."
					+ group.getChildTable();

			if (attributes.containsKey(nestedKey)) {
				List<Map<String, Object>> nestedValues = (List<Map<String, Object>>) attributes
						.get(nestedKey);

				Map<String, Entity> newEnclosingEntities = extendEntities(
						schema, table, enclosingEntities, thisEntity);

				for (Map<String, Object> nestedValue : nestedValues) {
					collectEntities(handle, group.getChildSchema(),
							group.getChildTable(), nestedValue, results,
							newEnclosingEntities, seq);
				}
			}
		}
	}

	private Map<String, Entity> extendEntities(String schema, String table,
			Map<String, Entity> enclosingEntities, Entity thisEntity) {
		Map<String, Entity> newEnclosingEntities = new LinkedHashMap<String, Entity>();
		newEnclosingEntities.putAll(enclosingEntities);
		newEnclosingEntities.put(SqlHelper.getTableName(schema, table),
				thisEntity);
		newEnclosingEntities = Collections
				.unmodifiableMap(newEnclosingEntities);
		return newEnclosingEntities;
	}
}