/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.dbo;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.StreamingOutput;

import org.skife.jdbi.v2.Handle;
import org.skife.jdbi.v2.IDBI;
import org.skife.jdbi.v2.Query;
import org.skife.jdbi.v2.Update;
import org.skife.jdbi.v2.tweak.HandleCallback;

import com.akiban.rest.model.ColumnInfo;
import com.akiban.rest.model.Entity;
import com.akiban.rest.model.GroupInfo;
import com.akiban.rest.model.PrimaryKeyInfo;
import com.akiban.rest.model.PseudoPk;
import com.akiban.rest.util.EncodingUtil;
import com.akiban.rest.util.StreamingJsonResult;
import com.google.common.collect.ImmutableMap;
import com.google.inject.Inject;

/**
 * Provides JDBC support for essential REST methods.
 */
public class JdbcDataSource {
	@Inject
	private IDBI database;

	@Inject
	private SqlPrimaryKeyHelper pkHelper;

	@Inject
	private SqlGroupHelper groupHelper;

	@Inject
	private SqlTableHelper tableHelper;

	@Inject
	private EntityHelper entityHelper;

	public Response execute(final String mode, final String jsonp,
			final Scanner sqlLines) {
		return Response.status(Status.OK)
				.entity(new StreamingJsonResult(mode, jsonp) {
					@Override
					public void doWrite() {
						database.withHandle(new HandleCallback<Void>() {
							@Override
							public Void withHandle(Handle handle)
									throws Exception {
								int line = 0;

								while (sqlLines.hasNextLine()) {
									line += 1;

									String trimmedLine = sqlLines.nextLine()
											.trim();
									if (trimmedLine.startsWith("--")
											|| trimmedLine.length() == 0) {
										continue;
									}

									try {
										handle.execute(trimmedLine);
										writeMapAsJson(ImmutableMap.of("line",
												line, "status", "OK", "sql",
												trimmedLine));
									} catch (Exception e) {
										writeMapAsJson(ImmutableMap.of("line",
												line, "status", "ERROR", "sql",
												trimmedLine, "message",
												e.getMessage()));
									}
								}

								return null;
							}
						});
					}
				}).build();
	}

	public Response retrieve(final String mode, final String jsonp,
			final String schema, final String table, final String identifiers,
			final Integer depth, final Integer offset, final Integer limit)
			throws Exception {
		boolean nullOffset = offset == null;
		boolean nullLimit = limit == null;
		if (nullLimit && !nullOffset) {
			// FIXME: investigate whether this is a bug in AK server
			return Response
					.status(Status.BAD_REQUEST)
					.entity("both offset and limit must be present when using offset")
					.build();
		}

		//
		// FIXME: method always returns 200 OK; we should modify the
		// code to handle more bad cases, like unknown table 404, bad input 400,
		// etc.
		//
		return Response.status(Status.OK)
				.entity(new StreamingJsonResult(mode, jsonp) {
					@Override
					public void doWrite() {
						HandleCallback<Void> handler = (identifiers == null || identifiers
								.length() < 1) ? getAllEntitiesHandler(schema,
								table, depth, offset, limit, this)
								: getSpecificEntitiesHandler(schema, table,
										identifiers, depth, this);

						database.withHandle(handler);
					}
				}).build();
	}

	public Response query(final String mode, final String jsonp,
			final String query, final Collection<?> parameters)
			throws Exception {
		//
		// FIXME : this verification could be better/more complete...
		//
		String trimmedLower = query != null ? query.toLowerCase().trim() : "";
		if (!trimmedLower.startsWith("select") || trimmedLower.contains(";")) {
			return Response
					.status(Status.BAD_REQUEST)
					.entity("query must begin with SELECT and contain only one statement")
					.build();
		}

		//
		// FIXME: method always returns 200 OK; we should modify the
		// code to handle unknown table 404, bad input 400, etc.
		//
		return Response.status(Status.OK)
				.entity(new StreamingJsonResult(mode, jsonp) {
					@Override
					public void doWrite() {
						database.withHandle(getQueryResultsHandler(this, query,
								parameters, false));
					}
				}).build();
	}

	public Response insert(final Handle handle, final String schema,
			final String table, final Map<String, Object> entity)
			throws Exception {
		//
		// FIXME : this is happy case only - we need to add error handling
		//
		if (entity.size() == 0) {
			return Response.status(Status.NOT_ACCEPTABLE).entity("").build();
		}

		List<GroupInfo> groupInfos = groupHelper.getParentGroupInfo(handle,
				schema, table);
		List<Entity> entities = entityHelper.parseEntities(handle, schema,
				table, entity, groupInfos);
		entities = populateSequenceValues(handle, entities);

		for (Entity theEntity : entities) {
			int result = performNonNestedInsert(handle, theEntity.getSchema(),
					theEntity.getTable(), theEntity.getAttributes());

			if (result < 1) {
				return Response.status(Status.NOT_ACCEPTABLE).entity("")
						.build();
			}
		}

		Entity rootEntity = entities.get(0);

		return Response
				.status(Status.OK)
				.header("Location", getLocation(rootEntity))
				.entity(EncodingUtil.toJson(Arrays.asList(rootEntity
						.getPrimaryKey()))).build();
	}

	public Response update(final Handle handle, final String schema,
			final String table, final String pk,
			final Map<String, Object> entity) throws Exception {
		//
		// FIXME : this is happy case only - we need to add error handling
		//
		List<Map<String, ?>> keys = pkHelper.parseIdentifierKeys(
				pkHelper.getPrimaryKeyInfo(handle, schema, table), pk);
		if (pk != null && pk.length() > 0 && keys.size() != 1) {
			return Response.status(Status.NOT_ACCEPTABLE).entity("").build();
		}

		List<GroupInfo> groupInfos = groupHelper.getParentGroupInfo(handle,
				schema, table);
		List<Entity> entities = entityHelper.parseEntities(handle, schema,
				table, entity, groupInfos);

		Map<String, ?> firstPk = keys.get(0);
		if (firstPk == null) {
			return Response.status(Status.NOT_ACCEPTABLE).entity("").build();
		}

		for (Entity theEntity : entities) {
			if (theEntity.getAttributes().size() == 0) {
				return Response.status(Status.NOT_ACCEPTABLE).entity("")
						.build();
			}

			for (Object object : theEntity.getPrimaryKey().values()) {
				if (object instanceof PseudoPk) {
					return Response.status(Status.NOT_ACCEPTABLE).entity("")
							.build();
				}
			}

			int result = performNonNestedUpdate(handle, theEntity.getSchema(),
					theEntity.getTable(), firstPk, theEntity);
			firstPk = null;

			if (result < 1) {
				return Response.status(Status.NOT_ACCEPTABLE).entity("")
						.build();
			}
		}

		return Response
				.status(Status.OK)
				.header("Location", getLocation(entities.get(0)))
				.entity(EncodingUtil.toJson(Arrays.asList(entities.get(0)
						.getPrimaryKey()))).build();
	}

	public Response delete(final String schema, final String table,
			final String identifiers) throws Exception {
		//
		// FIXME: method always returns 200 OK; we should modify the
		// code to handle unknown table 404, bad input 400, etc.
		//
		return Response.status(Status.OK).entity(new StreamingOutput() {
			@Override
			public void write(OutputStream out) throws IOException,
					WebApplicationException {
				final PrintWriter output = new PrintWriter(
						new OutputStreamWriter(out, "UTF-8"));
				try {
					HandleCallback<List<Map<String, ?>>> handler = deleteSpecificEntitiesHandler(
							schema, table, identifiers);

					List<Map<String, ?>> keys = database.withHandle(handler);

					output.println(EncodingUtil.toJson(keys));
				} finally {
					output.close();
				}
			}
		}).build();
	}

	private HandleCallback<Void> getAllEntitiesHandler(final String schema,
			final String table, final Integer depth, final Integer offset,
			final Integer limit, final StreamingJsonResult output) {
		return new HandleCallback<Void>() {
			@Override
			public Void withHandle(Handle handle) throws Exception {
				Iterator<Map<String, Object>> iterator = handle.createQuery(
						groupHelper.getRecursiveGroupSelect(handle, schema,
								table, depth, offset, limit)).iterator();

				SqlHelper.printResultSet(output, iterator, false);

				return null;
			}
		};
	}

	private HandleCallback<Void> getSpecificEntitiesHandler(
			final String schema, final String table, final String identifiers,
			final Integer depth, final StreamingJsonResult output) {
		return new HandleCallback<Void>() {
			public Void withHandle(Handle handle) throws Exception {
				PrimaryKeyInfo pkInfo = pkHelper.getPrimaryKeyInfo(handle,
						schema, table);

				List<Map<String, ?>> keys = pkHelper.parseIdentifierKeys(
						pkInfo, identifiers);

				Iterator<Map<String, ?>> keyIter = keys.iterator();
				while (keyIter.hasNext()) {
					Map<String, ?> key = keyIter.next();

					Query<Map<String, Object>> query = SqlHelper
							.getEntityQuery(groupHelper, handle, schema, table,
									key, depth);

					SqlHelper.printResultSet(output, query.iterator(), true);
				}

				return null;
			}
		};
	}

	public int performNonNestedUpdate(final Handle handle, final String schema,
			final String table, final Map<String, ?> key, final Entity entity)
			throws Exception {
		final ColumnInfo columns = tableHelper.getTableColumns(handle, schema,
				table);
		StringBuilder cols = new StringBuilder();
		boolean first = true;

		Map<String, Object> attributes = entity.getAttributes();

		Iterator<String> iter = columns.getColumns().iterator();
		while (iter.hasNext()) {
			String colName = iter.next();

			if (attributes.containsKey(colName)) {
				if (!first) {
					cols.append(", ");
				}

				cols.append(SqlHelper.quote(colName));
				cols.append(" = ?");
				first = false;
			}
		}

		if (first) {
			return 0;
		}

		StringBuilder updateStmt = new StringBuilder();
		updateStmt.append("update ");
		updateStmt.append(SqlHelper.getTableName(schema, table));
		updateStmt.append(" set ");
		updateStmt.append(cols);
		updateStmt.append(" where ");

		Map<String, ?> theKey = (key != null) ? key : entity.getPrimaryKey();

		first = true;
		for (Map.Entry<String, ?> entry : theKey.entrySet()) {
			if (!first) {
				updateStmt.append(" AND ");
			}

			updateStmt.append(SqlHelper.quote(entry.getKey()));
			updateStmt.append(" = ? ");
		}

		String query = updateStmt.toString();
		Update update = handle.createStatement(query);

		Iterator<String> iter2 = columns.getColumns().iterator();
		int i = 0;
		while (iter2.hasNext()) {
			String colName = iter2.next();
			if (attributes.containsKey(colName)) {
				update.bind(i, attributes.get(colName));
				i += 1;
			}
		}

		for (Map.Entry<String, ?> entry : theKey.entrySet()) {
			if (attributes.containsKey(entry.getKey())) {
				update.bind(i, entry.getValue());
				i += 1;
			}
		}

		return update.execute();
	}

	private HandleCallback<List<Map<String, ?>>> deleteSpecificEntitiesHandler(
			final String schema, final String table, final String identifiers) {
		return new HandleCallback<List<Map<String, ?>>>() {
			public List<Map<String, ?>> withHandle(Handle handle)
					throws Exception {
				PrimaryKeyInfo pkInfo = pkHelper.getPrimaryKeyInfo(handle,
						schema, table);

				List<Map<String, ?>> keys = pkHelper.parseIdentifierKeys(
						pkInfo, identifiers);

				//
				// FIXME : handle nested deletes based on group info
				//
				Iterator<Map<String, ?>> keyIter = keys.iterator();
				while (keyIter.hasNext()) {
					Map<String, ?> key = keyIter.next();

					Update update = SqlHelper.getEntityDeleteQuery(groupHelper,
							handle, schema, table, key);

					update.execute();
				}

				return keys;
			}
		};
	}

	public int performNonNestedInsert(final Handle handle, final String schema,
			final String table, final Map<String, Object> entity)
			throws Exception {
		final ColumnInfo columns = tableHelper.getTableColumns(handle, schema,
				table);
		StringBuilder cols = new StringBuilder();
		StringBuilder params = new StringBuilder();

		boolean first = true;

		Iterator<String> iter = columns.getColumns().iterator();
		while (iter.hasNext()) {
			String colName = iter.next();

			if (entity.containsKey(colName)) {
				if (!first) {
					cols.append(", ");
					params.append(", ");
				}

				cols.append(SqlHelper.quote(colName));
				params.append("?");
				first = false;
			}
		}

		StringBuilder insert = new StringBuilder();
		insert.append("insert into ");
		insert.append(SqlHelper.getTableName(schema, table));
		insert.append(" (");
		insert.append(cols);
		insert.append(" ) ");
		insert.append(" values (");
		insert.append(params);
		insert.append(")");

		String query = insert.toString();
		Update update = handle.createStatement(query);

		Iterator<String> iter2 = columns.getColumns().iterator();
		int i = 0;
		while (iter2.hasNext()) {
			String colName = iter2.next();
			if (entity.containsKey(colName)) {
				update.bind(i, entity.get(colName));
				i += 1;
			}
		}

		return update.execute();
	}

	private String getLocation(Entity entity) throws Exception {
		StringBuilder pathBuilder = new StringBuilder();
		pathBuilder.append("/");
		pathBuilder.append(URLEncoder.encode(entity.getSchema(), "UTF-8"));
		pathBuilder.append(".");
		pathBuilder.append(URLEncoder.encode(entity.getTable(), "UTF-8"));
		pathBuilder.append("/");

		boolean first = true;
		for (Map.Entry<String, Object> entry : entity.getPrimaryKey()
				.entrySet()) {
			if (!first) {
				pathBuilder.append(",");
			} else {
				first = false;
			}

			pathBuilder.append(URLEncoder.encode(entry.getKey(), "UTF-8"));
			pathBuilder.append("=");
			if (entry.getValue() != null) {
				pathBuilder.append(URLEncoder.encode(entry.getValue()
						.toString(), "UTF-8"));
			}
		}

		return pathBuilder.toString();
	}

	private List<Entity> populateSequenceValues(final Handle handle,
			final List<Entity> entities) throws Exception {
		Map<String, Object> seq = getSequenceValues(handle, entities);
		List<Entity> toReturn = new ArrayList<Entity>();

		for (Entity entity : entities) {
			Map<String, Object> newAttrs = new LinkedHashMap<String, Object>();

			for (Map.Entry<String, Object> entry : entity.getAttributes()
					.entrySet()) {
				if (entry.getValue() instanceof PseudoPk) {
					newAttrs.put(entry.getKey(),
							seq.get(((PseudoPk) (entry.getValue())).getKey()));
				} else {
					newAttrs.put(entry.getKey(), entry.getValue());
				}
			}

			for (Map.Entry<String, Object> entry : entity.getPrimaryKey()
					.entrySet()) {
				if (entry.getValue() instanceof PseudoPk) {
					newAttrs.put(entry.getKey(),
							seq.get(((PseudoPk) (entry.getValue())).getKey()));
				} else {
					newAttrs.put(entry.getKey(), entry.getValue());
				}
			}

			toReturn.add(new Entity(entity.getSchema(), entity.getTable(),
					entity.getPrimaryKeyInfo(), newAttrs, entity
							.getEnclosingEntities()));
		}

		return toReturn;
	}

	private Map<String, Object> getSequenceValues(final Handle handle,
			final List<Entity> entities) throws Exception {
		StringBuilder query = new StringBuilder();
		query.append("select ");
		boolean found = false;

		for (Entity entity : entities) {
			String key = null;

			if (entity.getPrimaryKeyInfo().isIdentitySequence()
					&& entity.getPrimaryKey().values().iterator().next() instanceof PseudoPk) {
				if (found) {
					query.append(", ");
				}

				found = true;
				key = ((PseudoPk) (entity.getPrimaryKey().values().iterator()
						.next())).getKey();
			} else {
				continue;
			}

			query.append(" (select next value for ");
			query.append(SqlHelper.getTableName(entity.getPrimaryKeyInfo()
					.getSequenceSchema(), entity.getPrimaryKeyInfo()
					.getSequenceName()));
			query.append(" as ");
			query.append(key);
			query.append(" ) as ");
			query.append(key);
		}

		if (!found) {
			return Collections.emptyMap();
		}

		Map<String, Object> result = handle.createQuery(query.toString())
				.first();
		String json = (String) result.get("json");
		Map<String, ?> sequenceResult = EncodingUtil.fromJson(
				LinkedHashMap.class, json);

		LinkedHashMap<String, Object> toReturn = new LinkedHashMap<String, Object>();

		for (Map.Entry<String, List<Map<String, Object>>> entry : ((Map<String, List<Map<String, Object>>>) sequenceResult)
				.entrySet()) {
			Map<String, Object> val = (Map<String, Object>) ((List) entry
					.getValue()).get(0);

			toReturn.put(entry.getKey(), val.get(entry.getKey()));
		}

		return toReturn;
	}

	private HandleCallback<Void> getQueryResultsHandler(
			final StreamingJsonResult output, final String query,
			final Collection<?> parameters, final boolean expectResult) {
		return new HandleCallback<Void>() {
			@Override
			public Void withHandle(Handle handle) throws Exception {
				Query queryObject = handle.createQuery(query);

				if (parameters != null && !parameters.isEmpty()) {
					int i = 0;
					for (Object parameter : parameters) {
						queryObject.bind(i, parameter);
						i += 1;
					}
				}

				SqlHelper.printResultSet(output, queryObject.iterator(),
						expectResult);

				return null;
			}
		};
	}
}
