/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.dbo;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.skife.jdbi.v2.Handle;

import com.akiban.rest.model.GroupInfo;
import com.akiban.rest.util.CacheHelper;
import com.akiban.rest.util.EncodingUtil;
import com.google.common.cache.Cache;

/**
 * Helper functions for working with table groups.
 */
public class SqlGroupHelper {
	private Cache<String, List<GroupInfo>> parentTableConfig = CacheHelper
			.getDefaultCache();

	private Cache<String, List<GroupInfo>> childTableConfig = CacheHelper
			.getDefaultCache();

	public String getRecursiveGroupSelect(Handle handle, String schema,
			String table, Integer depth, Integer offset, Integer limit)
			throws Exception {
		List<GroupInfo> groups = getParentGroupInfo(handle, schema, table);

		StringBuilder query = new StringBuilder();
		query.append("select *");

		if ((depth == null) || ((depth != null) && (depth >= 1))) {
			for (GroupInfo group : groups) {
				query.append(",\n");
				query.append("  (");
				query.append(getRecursiveGroupSelect(handle,
						group.getChildSchema(), group.getChildTable(),
						(depth != null) ? depth - 1 : null, null, null));
				query.append(" WHERE ");
				query.append(SqlHelper.getColumnName(group.getChildSchema(),
						group.getChildTable(), group.getLeftKey()));
				query.append(" = ");
				query.append(SqlHelper.getColumnName(group.getParentSchema(),
						group.getParentTable(), group.getRightKey()));
				query.append(") as `");
				query.append(group.getChildSchema());
				query.append(".");
				query.append(group.getChildTable());
				query.append("`");
			}
		}

		query.append(" from ");
		query.append(SqlHelper.getTableName(schema, table));

		if (limit != null) {
			query.append(" limit ");
			query.append(limit.toString());
		}

		if (offset != null) {
			query.append(" offset ");
			query.append(offset.toString());
		}

		return query.toString();
	}

	/**
	 * Retrieve group information from the information_schema where this table
	 * is a parent.
	 */
	public List<GroupInfo> getParentGroupInfo(final Handle handle,
			final String schema, final String table) throws Exception {
		String configKey = schema + "$$" + table;

		if (parentTableConfig.getIfPresent(configKey) == null) {
			String query = "SELECT "
					+ "  gc.unique_schema_name AS parent_schema, "
					+ "  gc.unique_table_name AS parent_table, "
					+ "  gc.constraint_schema_name AS child_schema, "
					+ "  gc.constraint_table_name AS child_table, "
					+ "  kcu.ordinal_position AS left_pos, "
					+ "  kcu.column_name AS left_key, "
					+ "  ic.ordinal_position AS right_pos, "
					+ "  ic.column_name AS right_key "
					+ "FROM "
					+ "  information_schema.grouping_constraints gc, "
					+ "  information_schema.key_column_usage kcu, "
					+ "  information_schema.index_columns ic "
					+ "WHERE kcu.constraint_name = gc.constraint_name"
					+ "  AND ic.schema_name = gc.unique_schema_name"
					+ "  AND ic.index_name = 'PRIMARY'"
					+ "  AND ic.column_table_name = gc.unique_table_name"
					+ "  AND gc.unique_schema_name = :schema "
					+ "  AND gc.unique_table_name = :table "
					+ "ORDER BY parent_table, child_table, left_pos, left_key, right_pos, right_key";

			List<GroupInfo> groupInfos = new ArrayList<GroupInfo>();
			List<Map<String, Object>> rows = handle.createQuery(query)
					.bind("schema", schema).bind("table", table).list();

			for (Map<String, Object> row : rows) {
				Map<String, Object> rowJson = EncodingUtil.fromJson(
						LinkedHashMap.class, (String) row.get("JSON"));

				groupInfos.add(new GroupInfo((String) rowJson
						.get("parent_schema"), (String) rowJson
						.get("parent_table"), (String) rowJson
						.get("child_schema"), (String) rowJson
						.get("child_table"), ((Number) rowJson.get("left_pos"))
						.intValue(), (String) rowJson.get("left_key"),
						((Number) rowJson.get("right_pos")).intValue(),
						(String) rowJson.get("right_key")));
			}

			parentTableConfig.put(configKey, groupInfos);
		}

		return parentTableConfig.getIfPresent(configKey);
	}

	/**
	 * Retrieve group information from the information_schema where this table
	 * is a child.
	 */
	public List<GroupInfo> getChildGroupInfo(final Handle handle,
			final String schema, final String table) throws Exception {
		String configKey = schema + "$$" + table;

		if (childTableConfig.getIfPresent(configKey) == null) {
			String query = "SELECT "
					+ "  gc.unique_schema_name AS parent_schema, "
					+ "  gc.unique_table_name AS parent_table, "
					+ "  gc.constraint_schema_name AS child_schema, "
					+ "  gc.constraint_table_name AS child_table, "
					+ "  kcu.ordinal_position AS left_pos, "
					+ "  kcu.column_name AS left_key, "
					+ "  ic.ordinal_position AS right_pos, "
					+ "  ic.column_name AS right_key "
					+ "FROM "
					+ "  information_schema.grouping_constraints gc, "
					+ "  information_schema.key_column_usage kcu, "
					+ "  information_schema.index_columns ic "
					+ "WHERE kcu.constraint_name = gc.constraint_name"
					+ "  AND ic.schema_name = gc.unique_schema_name"
					+ "  AND ic.index_name = 'PRIMARY'"
					+ "  AND ic.column_table_name = gc.unique_table_name "
					+ "  AND gc.constraint_schema_name = :schema "
					+ "  AND gc.constraint_table_name = :table "
					+ "ORDER BY parent_table, child_table, left_pos, left_key, right_pos, right_key";

			List<GroupInfo> groupInfos = new ArrayList<GroupInfo>();
			List<Map<String, Object>> rows = handle.createQuery(query)
					.bind("schema", schema).bind("table", table).list();

			for (Map<String, Object> row : rows) {
				Map<String, Object> rowJson = EncodingUtil.fromJson(
						LinkedHashMap.class, (String) row.get("JSON"));

				groupInfos.add(new GroupInfo((String) rowJson
						.get("parent_schema"), (String) rowJson
						.get("parent_table"), (String) rowJson
						.get("child_schema"), (String) rowJson
						.get("child_table"), ((Number) rowJson.get("left_pos"))
						.intValue(), (String) rowJson.get("left_key"),
						((Number) rowJson.get("right_pos")).intValue(),
						(String) rowJson.get("right_key")));
			}

			childTableConfig.put(configKey, groupInfos);
		}

		return childTableConfig.getIfPresent(configKey);
	}
}
