/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.dbo;

import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.skife.jdbi.v2.Handle;

import com.akiban.rest.model.PrimaryKeyInfo;
import com.akiban.rest.util.CacheHelper;
import com.akiban.rest.util.EncodingUtil;
import com.google.common.cache.Cache;

/**
 * Helper functions for working with primary keys.
 */
public class SqlPrimaryKeyHelper {
	private Cache<String, PrimaryKeyInfo> tableConfig = CacheHelper
			.getDefaultCache();

	/**
	 * Retrieve primary key information from the information_schema.
	 */
	public PrimaryKeyInfo getPrimaryKeyInfo(final Handle handle,
			final String schema, final String table) throws Exception {
		String configKey = schema + "$$" + table;

		if (tableConfig.getIfPresent(configKey) == null) {
			String query = "select ic.column_name, c.sequence_schema, c.sequence_name from "
					+ " information_schema.index_columns ic, "
					+ " information_schema.columns c "
					+ " where ic.index_name = 'PRIMARY' "
					+ "  and ic.schema_name = :schema "
					+ "  and ic.index_table_name = :table "
					+ "  and c.schema_name = :schema "
					+ "  and c.table_name = :table "
					+ "  and c.column_name = ic.column_name "
					+ " order by ic.ordinal_position";

			List<String> columns = new ArrayList<String>();

			List<Map<String, Object>> rows = handle.createQuery(query)
					.bind("schema", schema).bind("table", table).list();

			if (rows.size() < 1) {
				throw new IllegalArgumentException("Table has no primary key!");
			}

			String sequenceSchema = null;
			String sequenceName = null;

			for (Map<String, Object> row : rows) {
				Map<String, Object> rowJson = EncodingUtil.fromJson(
						LinkedHashMap.class, (String) row.get("JSON"));

				columns.add((String) rowJson.get("column_name"));

				sequenceSchema = (String) rowJson.get("sequence_schema");
				sequenceName = (String) rowJson.get("sequence_name");
			}

			PrimaryKeyInfo pkInfo = new PrimaryKeyInfo(columns, sequenceSchema,
					sequenceName);

			tableConfig.put(configKey, pkInfo);
		}

		return tableConfig.getIfPresent(configKey);
	}

	/**
	 * Parse identifier keys: format is (1) semicolon-delimited list of keys
	 * which are in (2) comma-delimited list of (3) URL-encoded fields or (4)
	 * column/value pair separated by "="
	 */
	public List<Map<String, ?>> parseIdentifierKeys(PrimaryKeyInfo pkInfo,
			String identifiers) {
		//
		// FIXME: make this a lot more bulletproof
		//
		List<Map<String, ?>> results = new ArrayList<Map<String, ?>>();
		String[] ids = identifiers.split(";");
		List<String> columns = pkInfo.getColumns();
		int keyLen = columns.size();

		for (String id : ids) {
			Map<String, Object> key = new LinkedHashMap<String, Object>();
			String[] kvPairs = id.split(",");

			if (kvPairs.length != keyLen) {
				throw new IllegalArgumentException(
						"Primary key column count mismatch: " + kvPairs);
			}

			boolean hasFully = false;
			for (int i = 0; i < keyLen; i++) {
				String kvPair = kvPairs[i];
				String defaultKey = columns.get(i);
				String value = kvPair;

				String[] kv = kvPair.split("=");
				if (kv.length == 2) {
					hasFully = true;
					defaultKey = kv[0].trim();
					value = kv[1].trim();
				} else if (hasFully) {
					throw new IllegalArgumentException(
							"If one component is qualified, all key components must be qualified: "
									+ kvPair);
				}

				if (!columns.contains(defaultKey)) {
					throw new IllegalArgumentException(
							"Only primary key columns may be specified: "
									+ kvPair);
				}

				Object valueObject = URLDecoder.decode(value);
				if (keyLen == 1 && pkInfo.isIdentitySequence()) {
					valueObject = Long.parseLong(valueObject.toString());
				}

				key.put(defaultKey, valueObject);
			}

			results.add(key);
		}

		return results;
	}
}