/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.model;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.akiban.rest.util.EncodingUtil;

/**
 * Encapsulates information about a table's primary key.
 */
public class PrimaryKeyInfo {
	private final List<String> columns;
	private boolean identitySequence;
	private final String sequenceSchema;
	private final String sequenceName;

	public PrimaryKeyInfo(List<String> columns, String sequenceSchema,
			String sequenceName) {
		this.columns = Collections.unmodifiableList(columns);
		this.identitySequence = sequenceSchema != null && sequenceName != null;
		this.sequenceSchema = sequenceSchema;
		this.sequenceName = sequenceName;
	}

	public List<String> getColumns() {
		return columns;
	}

	public String getSequenceName() {
		return sequenceName;
	}

	public String getSequenceSchema() {
		return sequenceSchema;
	}

	public boolean isIdentitySequence() {
		return identitySequence;
	}

	public boolean hasPrimaryKey(Map<String, Object> entity) {
		for (String column : columns) {
			if (!entity.containsKey(column) || entity.get(column) == null) {
				return false;
			}
		}

		return true;
	}

	public Map<String, Object> getPrimaryKey(Map<String, Object> attributes) {
		Map<String, Object> primaryKey = new LinkedHashMap<String, Object>();

		for (String column : columns) {
			primaryKey.put(column, attributes.get(column));
		}

		return primaryKey;
	}

	@Override
	public String toString() {
		return "PrimaryKeyInfo:" + EncodingUtil.toJson(this);
	}
}