/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.util;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Map;

import javax.ws.rs.core.StreamingOutput;

/**
 * Encapsulates the output of Streaming JSON result sets.
 */
public abstract class StreamingJsonResult implements StreamingOutput {
	public enum OutputMode {
		FLAT, ARRAY, JSONP
	}

	protected final OutputMode mode;
	protected final String jsonp;
	protected PrintWriter output;
	protected boolean first = true;

	public StreamingJsonResult(String mode, String jsonp) {
		this.mode = (mode != null) ? OutputMode.valueOf(mode.toUpperCase())
				: ((jsonp != null) ? OutputMode.JSONP : OutputMode.ARRAY);
		this.jsonp = (jsonp != null) ? jsonp : "";
	}

	public void writeJsonString(String json) {
		boolean wasFirst = first;

		if (first) {
			writeHeader();
			first = false;
		} else if (!OutputMode.FLAT.equals(mode)) {
			output.print(",");
		}

		if (!wasFirst || !OutputMode.FLAT.equals(mode)) {
			output.println();
		}

		output.print(json);
		output.flush();
	}

	public void writeMapAsJson(Map<String, ?> jsonMap) {
		try {
			writeJsonString(EncodingUtil.toJson(jsonMap));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public abstract void doWrite();

	@Override
	public void write(OutputStream outputStream) {
		this.output = new PrintWriter(outputStream);
		try {
			doWrite();
		} finally {
			close();
		}
	}

	private void writeHeader() {
		// NOTE: fallthrough is intentional
		switch (mode) {
		case JSONP:
			output.print(jsonp + "(");
		case ARRAY:
			output.print("[");
		}
	}

	private void writeFooter() {
		if (first) {
			writeHeader();
		}

		if (!OutputMode.FLAT.equals(mode) && !first) {
			output.println();
		}

		// NOTE: fallthrough is intentional
		switch (mode) {
		case JSONP:
		case ARRAY:
			output.print("]");
		}

		if (mode.equals(OutputMode.JSONP)) {
			output.print(");");
		}

		output.println();
	}

	private void close() {
		writeFooter();
		output.close();
	}
}