/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;

import org.testng.Assert;
import org.testng.annotations.Test;

import com.akiban.rest.util.EncodingUtil;
import com.google.common.collect.ImmutableMap;

/**
 * Test cases for the DataAccessOperationsResource - essentially making sure
 * that all the Create Retrieve Update and Delete functions work.
 */
@Test
public class DataAccessOperationsResourceTest {
	private SqlExecutionResource exec = TestHelper
			.get(SqlExecutionResource.class);
	private DataAccessOperationsResource dao = TestHelper
			.get(DataAccessOperationsResource.class);
	private static final String FLAT_SCHEMA = "t20121113";
	private static final String FLAT_TABLE = "table_01";
	private static final String NEST_SCHEMA = "g_coi";
	private static final String NEST_TABLE = "customers";

	public void testGetAll() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertEquals(getAll(FLAT_SCHEMA, FLAT_TABLE),
				Collections.emptyList());

		Map<String, ?> posted = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted, ImmutableMap.of("a", 1));

		Assert.assertEquals(getAll(FLAT_SCHEMA, FLAT_TABLE),
				Arrays.asList(ImmutableMap.of("a", 1)));
	}

	public void testGetSingle() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertNull(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"));

		Map<String, ?> posted = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted, ImmutableMap.of("a", 1));

		Assert.assertEquals(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"),
				ImmutableMap.of("a", 1));
	}

	public void testGetMulti() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertEquals(getMulti(FLAT_SCHEMA, FLAT_TABLE, "1", "2"),
				Arrays.asList(null, null));

		Map<String, ?> posted1 = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted1, ImmutableMap.of("a", 1));

		Assert.assertEquals(getMulti(FLAT_SCHEMA, FLAT_TABLE, "1", "2"),
				Arrays.asList(ImmutableMap.of("a", 1), null));

		Map<String, ?> posted2 = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted2, ImmutableMap.of("a", 2));

		Assert.assertEquals(getMulti(FLAT_SCHEMA, FLAT_TABLE, "1", "2"),
				Arrays.asList(ImmutableMap.of("a", 1), ImmutableMap.of("a", 2)));
		Assert.assertEquals(getMulti(FLAT_SCHEMA, FLAT_TABLE, "2", "1"),
				Arrays.asList(ImmutableMap.of("a", 2), ImmutableMap.of("a", 1)));
	}

	public void testPostFlat() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertNull(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"));

		Map<String, ?> posted = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted, ImmutableMap.of("a", 1));

		Assert.assertEquals(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"),
				ImmutableMap.of("a", 1));
	}

	public void testPutFlat() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertNull(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"));

		Map<String, ?> posted = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted, ImmutableMap.of("a", 1));

		Assert.assertEquals(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"),
				ImmutableMap.of("a", 1));

		Map<String, ?> put = putSingle(FLAT_SCHEMA, FLAT_TABLE, "1",
				EncodingUtil.fromJson(Map.class, "{\"a\":17}"));
		Assert.assertEquals(put, ImmutableMap.of("a", 17));

		Assert.assertEquals(getSingle(FLAT_SCHEMA, FLAT_TABLE, "17"),
				ImmutableMap.of("a", 17));
	}

	public void testDeleteFlat() {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Assert.assertNull(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"));

		Map<String, ?> posted = postSingle(FLAT_SCHEMA, FLAT_TABLE,
				EncodingUtil.fromJson(Map.class, "{\"a\":null}"));
		Assert.assertEquals(posted, ImmutableMap.of("a", 1));

		Assert.assertEquals(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"),
				ImmutableMap.of("a", 1));

		Map<String, ?> del = deleteSingle(FLAT_SCHEMA, FLAT_TABLE, "1");
		Assert.assertEquals(del, ImmutableMap.of("a", 1));

		Assert.assertNull(getSingle(FLAT_SCHEMA, FLAT_TABLE, "1"));

		Map<String, ?> delNoExist = deleteSingle(FLAT_SCHEMA, FLAT_TABLE, "1");
		Assert.assertEquals(delNoExist, ImmutableMap.of("a", 1));
	}

	public void testPostNested() {
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_create.sql");
		Assert.assertNull(getSingle(NEST_SCHEMA, NEST_TABLE, "3"));

		Map<String, Object> customer = getCustomer();
		Map<String, Object> address = getAddress();
		Map<String, Object> order = getOrder();
		Map<String, Object> item = getItem();

		order.put(NEST_SCHEMA + ".items", Arrays.asList(item));
		customer.put(NEST_SCHEMA + ".addresses", Arrays.asList(address));
		customer.put(NEST_SCHEMA + ".orders", Arrays.asList(order));

		Map<String, ?> posted = postSingle(NEST_SCHEMA, NEST_TABLE, customer);
		Assert.assertEquals(posted, ImmutableMap.of("cid", 3));

		customer.put("cid", 3);
		address.put("aid", 2);
		address.put("cid", 3);
		order.put("cid", 3);
		order.put("oid", 1);
		item.put("oid", 1);
		item.put("iid", 63);

		Map<String, ?> result = getSingle(NEST_SCHEMA, NEST_TABLE, "3");
		Assert.assertEquals(result.toString(), customer.toString());
	}

	public void testPutNested() {
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_create.sql");

		Map<String, Object> customer = (Map<String, Object>) getSingle(
				NEST_SCHEMA, NEST_TABLE, "1");
		Assert.assertNotNull(customer);

		Map<String, Object> order = ((List<Map<String, Object>>) customer
				.get("g_coi.orders")).get(0);

		customer.put("first_name", "zzz");
		order.put("ship_method", "zzz");

		Map<String, ?> posted = putSingle(NEST_SCHEMA, NEST_TABLE, "1",
				customer);

		Assert.assertEquals(posted, ImmutableMap.of("cid", 1));

		Map<String, ?> modified = getSingle(NEST_SCHEMA, NEST_TABLE, "1");
		Assert.assertEquals(modified.toString(), customer.toString());
		Assert.assertEquals(modified.get("first_name"), "zzz");
		Assert.assertEquals(
				((List<Map<String, Object>>) modified.get("g_coi.orders")).get(
						0).get("ship_method"), "zzz");
	}

	private List<Map<String, ?>> getAll(String schema, String table) {
		try {
			Response response = dao.retrieveEntity(null, null, schema, table,
					0, 0, 10);

			return TestHelper.getObject(List.class, response);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private List<Map<String, ?>> getMulti(String schema, String table,
			String... ids) {
		try {
			Response response = dao.retrieveEntity(null, null, schema, table,
					null, getPathUri(ids));
			List<Map<String, ?>> result = TestHelper.getObject(List.class,
					response);
			Assert.assertEquals(result.size(), ids.length);

			return result;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private Map<String, ?> getSingle(String schema, String table, String id) {
		try {
			Response response = dao.retrieveEntity(null, null, schema, table,
					null, getPathUri(id));

			List<Map<String, ?>> result = TestHelper.getObject(List.class,
					response);
			Assert.assertEquals(result.size(), 1);

			return result.get(0);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private Map<String, ?> postSingle(String schema, String table,
			Map<String, ?> instance) {
		try {
			Response response = dao.createEntity(schema, table, EncodingUtil
					.toJson(instance).getBytes());

			return (Map<String, ?>) TestHelper.getObject(List.class, response)
					.get(0);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private Map<String, ?> putSingle(String schema, String table, String id,
			Map<String, ?> instance) {
		try {
			Response response = dao.updateEntity(schema, table, EncodingUtil
					.toJson(instance).getBytes(), getPathUri(id));

			return (Map<String, ?>) TestHelper.getObject(List.class, response)
					.get(0);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private Map<String, ?> deleteSingle(String schema, String table, String id) {
		try {
			Response response = dao.deleteEntity(schema, table, getPathUri(id));

			return (Map<String, ?>) TestHelper.getObject(List.class, response)
					.get(0);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private List<Map<String, ?>> deleteMulti(String schema, String table,
			String... ids) {
		try {
			Response response = dao
					.deleteEntity(schema, table, getPathUri(ids));

			return TestHelper.getObject(List.class, response);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private static UriInfo getPathUri(String... ids) {
		UriInfo info = TestHelper.getUri(join("/", "", join(";", ids)));

		return info;
	}

	private static String join(String delim, String... parts) {
		if (parts == null || parts.length == 0) {
			return "";
		}

		StringBuilder builder = new StringBuilder();
		for (String part : parts) {
			builder.append(delim);
			builder.append(part);
		}

		return builder.toString().substring(delim.length());
	}

	private static Map<String, Object> getCustomer() {
		Map<String, Object> toPost = new LinkedHashMap<String, Object>();
		toPost.put("cid", null);
		toPost.put("first_name", "firstName");
		toPost.put("last_name", "lastName");
		toPost.put("primary_payment_code", "1234");
		toPost.put("payment_status", "Y");
		toPost.put("comment", "aComment");

		return toPost;
	}

	private static Map<String, Object> getOrder() {
		Map<String, Object> toPost = new LinkedHashMap<String, Object>();
		toPost.put("oid", null);
		toPost.put("cid", null);
		toPost.put("order_date", "2012-11-01");
		toPost.put("order_status", "S");
		toPost.put("order_priority", "1");
		toPost.put("ship_priority", "1");
		toPost.put("ship_method", "A");
		toPost.put("ship_parts", 1);

		return toPost;
	}

	private static Map<String, Object> getItem() {
		Map<String, Object> toPost = new LinkedHashMap<String, Object>();
		toPost.put("iid", null);
		toPost.put("oid", null);
		toPost.put("sku", 12345);
		toPost.put("quantity", 10);
		toPost.put("unit_price", "1.00");
		toPost.put("discount", "0.5");
		toPost.put("tax", "0.05");
		toPost.put("item_status", "A");
		toPost.put("shipment_part", 1);
		toPost.put("color", "red");

		return toPost;
	}

	private static Map<String, Object> getAddress() {
		Map<String, Object> toPost = new LinkedHashMap<String, Object>();
		toPost.put("aid", null);
		toPost.put("cid", null);
		toPost.put("state", "AK");
		toPost.put("zip_code", 12345);
		toPost.put("phone", "123-456-7890");
		toPost.put("address_type", "B");

		return toPost;
	}
}
