/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.ws.rs.core.Response;

import org.testng.Assert;
import org.testng.annotations.Test;

/**
 * Test cases for the SqlGroupsResource - essentially making sure that the
 * retrieval of table group information works.
 */
@Test
public class SqlGroupsResourceTest {
	private SqlExecutionResource exec = TestHelper
			.get(SqlExecutionResource.class);
	private SqlGroupsResource groups = TestHelper.get(SqlGroupsResource.class);
	private static final String SCHEMA = "g_coi";

	public void testEmptyGroupsNoSchema() throws Exception {
		List<?> hopefullyEmpty = getTableNames("random2189012912");
		Assert.assertTrue(hopefullyEmpty.isEmpty());
	}

	public void testFlatGroups() throws Exception {
		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_drop.sql");
		String testSchema = "t20121113";

		Response beforeResponse = groups.getGroups(testSchema, null, null);
		Assert.assertTrue(TestHelper.getObject(List.class, beforeResponse)
				.isEmpty());

		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_create.sql");
		Response afterResponse = groups.getGroups(testSchema, null, null);
		List<Map<String, ?>> flatGroups = TestHelper.getObject(List.class,
				afterResponse);
		List<?> tableDepths = TestHelper.collectList(flatGroups, "depth");

		Assert.assertEquals(tableDepths.size(), 1);
		for (Object object : tableDepths) {
			Assert.assertEquals(object.toString(), "0");
		}

		TestHelper.executeSql(exec, "com/akiban/rest/exec_01_drop.sql");
	}

	public void testNestedGroups() throws Exception {
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");

		Response beforeResponse = groups.getGroups(SCHEMA, null, null);
		Assert.assertTrue(TestHelper.getObject(List.class, beforeResponse)
				.isEmpty());

		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_create.sql");
		Response afterResponse = groups.getGroups(SCHEMA, null, null);
		List<Map<String, ?>> flatGroups = TestHelper.getObject(List.class,
				afterResponse);

		for (Map<String, ?> entry : flatGroups) {
			Assert.assertEquals(entry.get("root_schema_name"), SCHEMA);
			Assert.assertEquals(entry.get("root_table_name"), "customers");
			Assert.assertEquals(entry.get("constraint_schema_name"), SCHEMA);
		}

		Assert.assertEquals(TestHelper.collectList(flatGroups, "depth"),
				Arrays.asList(0, 1, 1, 2));

		Assert.assertEquals(
				TestHelper.collectList(flatGroups, "constraint_table_name"),
				Arrays.asList("customers", "addresses", "orders", "items"));

		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");
	}

	private List<?> getTableNames(String schema) throws Exception {
		Response response = groups.getGroups(schema, null, null);
		List<?> list = TestHelper.collectList(
				TestHelper.getObject(List.class, response),
				"constraint_table_name");

		return list;
	}
}
