/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.util.List;
import java.util.Map;

import javax.ws.rs.core.Response;

import org.testng.Assert;
import org.testng.annotations.Test;

import com.google.common.collect.ImmutableMap;

/**
 * Test cases for the SqlTablesResource - ensuring that retrieval of table
 * information and details works properly.
 */
@Test
public class SqlTablesResourceTest {
	private SqlExecutionResource exec = TestHelper
			.get(SqlExecutionResource.class);
	private SqlTablesResource tables = TestHelper.get(SqlTablesResource.class);
	private static final String SCHEMA = "g_coi";

	public void testEmptyTableList() throws Exception {
		List<?> hopefullyEmpty = getTableNames("random2189012912");
		Assert.assertTrue(hopefullyEmpty.isEmpty());
	}

	public void testTableList() throws Exception {
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");
		List<?> listBefore = getTableNames(SCHEMA);

		Assert.assertTrue(listBefore.isEmpty());

		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_create.sql");
		List<?> listAfter = getTableNames(SCHEMA);

		Assert.assertTrue(listAfter.contains("items"));
		Assert.assertTrue(listAfter.contains("orders"));
		Assert.assertTrue(listAfter.contains("addresses"));
		Assert.assertTrue(listAfter.contains("customers"));
		Assert.assertFalse(listAfter.contains("nonexistent"));

		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_drop.sql");

		List<?> listAfterDrop = getTableNames(SCHEMA);

		Assert.assertTrue(listAfterDrop.isEmpty());
	}

	public void testEmptyTableDetails() throws Exception {
		List<Map<String, ?>> schemaNonexistent = TestHelper.getObject(
				List.class, tables.getTableDetails(
						"random2189012912.random2189012912", null, null));

		Assert.assertTrue(schemaNonexistent.isEmpty());

		List<Map<String, ?>> tableNonexistent = TestHelper.getObject(
				List.class, tables.getTableDetails(
						"information_schema.random2189012912", null, null));

		Assert.assertTrue(tableNonexistent.isEmpty());
	}

	public void testTableDetails() throws Exception {
		TestHelper.executeSql(exec, "com/akiban/rest/groups_01_create.sql");
		String tableName = "customers";

		Response detailsResponse = tables.getTableDetails(SCHEMA + "."
				+ tableName, null, null);
		Map<String, Object> details = (Map<String, Object>) TestHelper
				.getObject(List.class, detailsResponse).get(0);

		TestHelper.assertMatch(details, ImmutableMap.<String, Object> of(
				"table_schema", SCHEMA, "table_name", tableName, "table_type",
				"TABLE"));

		List<?> columnNames = TestHelper.collectList(
				(List<Map<String, ?>>) details.get("columns"), "column_name");

		Assert.assertTrue(columnNames.contains("cid"));
		Assert.assertTrue(columnNames.contains("first_name"));
		Assert.assertTrue(columnNames.contains("last_name"));
		Assert.assertTrue(columnNames.contains("primary_payment_code"));
		Assert.assertTrue(columnNames.contains("payment_status"));
		Assert.assertTrue(columnNames.contains("comment"));
		Assert.assertFalse(columnNames.contains("nonexistent"));

		List<?> indexNames = TestHelper.collectList(
				(List<Map<String, ?>>) details.get("indexes"), "index_name");

		Assert.assertTrue(indexNames.contains("PRIMARY"));
		Assert.assertFalse(indexNames.contains("nonexistent"));

		List<?> indexColumnNames = TestHelper.collectList(
				(List<Map<String, ?>>) ((List<Map<String, ?>>) details
						.get("indexes")).get(0).get("index_columns"),
				"column_name");

		Assert.assertTrue(indexColumnNames.contains("cid"));
		Assert.assertFalse(indexColumnNames.contains("nonexistent"));

		List<?> constraintNames = TestHelper.collectList(
				(List<Map<String, ?>>) details.get("table_constraints"),
				"constraint_name");

		Assert.assertTrue(constraintNames.contains("PRIMARY"));
		Assert.assertFalse(constraintNames.contains("nonexistent"));
	}

	private List<?> getTableNames(String schema) throws Exception {
		Response response = tables.getTableDetails(schema, null, null);
		List<?> list = TestHelper.collectList(
				TestHelper.getObject(List.class, response), "table_name");

		return list;
	}
}
