/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.ws.rs.core.Response;
import javax.ws.rs.core.StreamingOutput;
import javax.ws.rs.core.UriInfo;

import org.testng.Assert;

import com.akiban.rest.test.util.UriInfoBase;
import com.akiban.rest.util.EncodingUtil;
import com.google.common.collect.ImmutableMap;
import com.google.inject.Guice;
import com.google.inject.Injector;

/**
 * Essential functions for REST service testing.
 */
public class TestHelper {
	private static final Injector INJECTOR;
	static {
		INJECTOR = Guice.createInjector(new Main.EmbeddedRestModule(),
				new RestServiceModule());
		INJECTOR.getInstance(RestServiceImpl.class).start();
	}

	public static Injector getInjector() {
		return INJECTOR;
	}

	public static <T> T get(Class<T> clazz) {
		return INJECTOR.getInstance(clazz);
	}

	public static void executeSql(SqlExecutionResource exec, String sqlResource) {
		try {
			TestHelper.getObject(
					List.class,
					exec.execute(null, null,
							TestHelper.getResource(sqlResource)));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public static UriInfo getUri(final String path) {
		return new UriInfoBase() {
			@Override
			public String getPath() {
				return path;
			}
		};
	}

	public static InputStream getResource(String name) {
		return TestHelper.class.getClassLoader().getResourceAsStream(name);
	}

	public static <T> T getObject(Class<T> clazz, Response response) {
		return EncodingUtil.fromJson(clazz, getString(response));
	}

	public static void assertMatch(Response response, int statusCode,
			Map<String, String> headers, String body) {
		Assert.assertEquals(response.getStatus(), statusCode);

		for (Map.Entry<String, String> header : headers.entrySet()) {
			Assert.assertEquals(response.getMetadata()
					.getFirst(header.getKey()), header.getValue());
		}

		Assert.assertEquals(response.getEntity().toString(), body);
	}

	public static void assertMatch(Map<String, ?> actual,
			Map<String, ?> expected) {
		for (Map.Entry<String, ?> entry : expected.entrySet()) {
			Assert.assertEquals(actual.get(entry.getKey()), entry.getValue());
		}
	}

	public static List<Map<String, ?>> collectMap(List<Map<String, ?>> result,
			String... columns) {
		List<Map<String, ?>> toReturn = new ArrayList<Map<String, ?>>();

		for (Map<String, ?> value : result) {
			ImmutableMap.Builder<String, Object> builder = ImmutableMap
					.builder();
			for (String column : columns) {
				builder.put(column, value.get(column));
			}

			toReturn.add(builder.build());
		}

		return toReturn;
	}

	public static List<?> collectList(List<Map<String, ?>> result, String column) {
		List<Object> toReturn = new ArrayList<Object>();

		for (Map<String, ?> value : result) {
			toReturn.add((Object) value.get(column));
		}

		return toReturn;
	}

	private static String getString(Response response) {
		Object entity = response.getEntity();

		if (entity instanceof String) {
			return (String) entity;
		}

		return new String(getBytes(response));
	}

	private static byte[] getBytes(Response response) {
		Object entity = response.getEntity();

		if (entity instanceof String) {
			return ((String) entity).getBytes();
		}

		try {
			StreamingOutput json = (StreamingOutput) entity;
			ByteArrayOutputStream output = new ByteArrayOutputStream();
			json.write(output);

			return output.toByteArray();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
}
