/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.util;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.concurrent.atomic.AtomicLong;

import javax.sql.DataSource;

import org.testng.Assert;
import org.testng.annotations.Test;

import com.akiban.rest.RestServiceImpl;
import com.akiban.rest.TestHelper;
import com.jolbox.bonecp.BoneCPDataSource;

/**
 * Test case for the BoneCP data source - ensuring it connects properly and
 * reconnects on failure.
 */
@Test
public class DataSourceTest {
	public void testValidConnection() throws Exception {
		BoneCPDataSource ds = getDataSource();
		Connection conn = ds.getConnection();

		doSimpleSelect(conn, false);
		doSimpleSelect(conn, true);

		try {
			doSimpleSelect(conn, true);
			Assert.fail("expected exception");
		} catch (SQLException expected) {
			Assert.assertTrue(expected.getMessage().startsWith(
					"Connection is closed!"));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public void testInvalidConnection() throws Exception {
		System.setProperty("plugins.rest.jdbc.url",
				"jdbc:postgresql://127.0.0.1:12345/");

		try {
			BoneCPDataSource ds = getDataSource();
			try {
				doSimpleSelect(ds.getConnection(), true);
				Assert.fail("expected exception");
			} catch (SQLException expected) {
				Assert.assertTrue(expected.getMessage().startsWith(
						"Unable to open a test connection"));
			} catch (Exception e) {
				throw new RuntimeException(e);
			}

			ds.setJdbcUrl("jdbc:postgresql://127.0.0.1:15432/");
			doSimpleSelect(ds.getConnection(), true);
		} finally {
			System.clearProperty("plugins.rest.jdbc.url");
		}
	}

	public void testReconnection() throws Exception {
		DataSource ds = wrapProxy((BoneCPDataSource) getDataSource());
		doSimpleSelect(ds.getConnection(), true);

		try {
			doSimpleSelect(ds.getConnection(), true);
			Assert.fail("expected exception");
		} catch (SQLException expected) {
			Assert.assertTrue(expected.getMessage().startsWith(
					"Unable to open a test connection"));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		doSimpleSelect(ds.getConnection(), true);
	}

	private void doSimpleSelect(Connection conn, boolean close)
			throws SQLException {
		try {
			PreparedStatement stmt = conn.prepareStatement("select 1 as good");
			ResultSet rs = stmt.executeQuery();
			Assert.assertTrue(rs.next());
			Assert.assertEquals(rs.getString(1), "{\"good\":1}");
		} finally {
			if (conn != null && close) {
				conn.close();
			}
		}
	}

	private DataSource wrapProxy(final BoneCPDataSource ds) {
		ds.setAcquireRetryAttempts(0);

		return (DataSource) Proxy.newProxyInstance(getClass().getClassLoader(),
				new Class[] { DataSource.class }, new InvocationHandler() {
					private AtomicLong count = new AtomicLong();

					@Override
					public Object invoke(Object instance, Method method,
							Object[] args) throws Throwable {
						if (method.getName().equals("getConnection")) {
							long i = count.getAndIncrement();
							if (i == 1) {
								throw new SQLException(
										"Unable to open a test connection");
							}
						}

						try {
							return method.invoke(ds, args);
						} catch (InvocationTargetException e) {
							throw e.getCause();
						}
					}
				});
	}

	private BoneCPDataSource getDataSource() {
		Props props = TestHelper.get(Props.class);
		BoneCPDataSource dataSource = new BoneCPDataSource();
		RestServiceImpl.configureDataSource(dataSource, props);

		return dataSource;
	}
}
