/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akiban.rest.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.codehaus.jackson.JsonGenerationException;
import org.codehaus.jackson.map.JsonMappingException;
import org.testng.Assert;
import org.testng.annotations.Test;

import com.google.common.collect.ImmutableMap;

/**
 * Test cases for the StreamingJsonResult - ensuring that FLAT, JSONP, and ARRAY
 * formats work properly.
 */
@Test
public class StreamingJsonResultTest {
	public void testCompletelyEmptyResults() {
		performJsonResultTests();
	}

	public void testSingleObjectEmptyResults() {
		performJsonResultTests((Object) null);
		performJsonResultTests("");
		performJsonResultTests(Collections.emptyList());
		performJsonResultTests(Collections.emptyMap());
	}

	public void testSingleObjectPrimitiveResults() {
		performJsonResultTests(101);
		performJsonResultTests("it works");
		performJsonResultTests(false);
	}

	public void testSingleObjectMapResults() {
		performJsonResultTests(ImmutableMap.<String, Object> of("foo", 1,
				"bar", "bar value", "baz", true));
	}

	public void testNoResultsPrinted() {
		StreamingJsonResult result = new StreamingJsonResult("array", null) {
			@Override
			public void doWrite() {
			}
		};

		ByteArrayOutputStream output = new ByteArrayOutputStream();
		result.write(output);

		String resultString = new String(output.toByteArray()).replaceAll("\n",
				"");

		Assert.assertEquals(resultString, "[]");
	}

	private static void performJsonResultTests(Object... values) {
		performFlatTest(Arrays.asList(values));
		performArrayTest(Arrays.asList(values));
		performJsonpTest(Arrays.asList(values));
	}

	private static void performFlatTest(final List<Object> values) {
		String resultString = getResultString("flat", null, values);
		try {
			assertMatch(resultString, values);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private static void performArrayTest(final List<Object> values) {
		String resultString = getResultString("array", null, values);
		String prefix = "[";
		String suffix = "]";

		Assert.assertTrue(resultString.startsWith(prefix));

		String toMatch = resultString.substring(prefix.length(),
				resultString.length() - suffix.length());
		try {
			assertMatch(toMatch, values);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		Assert.assertTrue(resultString.endsWith(suffix));
	}

	private static void performJsonpTest(final List<Object> values) {
		String callback = "doTest";
		String resultString = getResultString("jsonp", callback, values);
		String prefix = callback + "([";
		String suffix = "]);";

		Assert.assertTrue(resultString.startsWith(prefix));

		String toMatch = resultString.substring(prefix.length(),
				resultString.length() - suffix.length());
		try {
			assertMatch(toMatch, values);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		Assert.assertTrue(resultString.endsWith(suffix));
	}

	private static String getResultString(String mode, final String callback,
			final List<Object> values) {
		StreamingJsonResult result = new StreamingJsonResult(mode, callback) {
			@Override
			public void doWrite() {
				try {
					for (Object value : values) {
						writeJsonString(EncodingUtil.toJson(value));
					}
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
			}
		};

		ByteArrayOutputStream output = new ByteArrayOutputStream();
		result.write(output);

		return new String(output.toByteArray()).replaceAll("\n", "");
	}

	private static void assertMatch(String resultString,
			final List<Object> values) throws IOException,
			JsonGenerationException, JsonMappingException {
		String result = EncodingUtil.toJson(values).replaceAll("\n", "");
		Assert.assertEquals(resultString,
				result.substring(1, result.length() - 1));
	}
}